﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Collections;
using HeuristicLab.Core;
using HeuristicLab.Core.Views;
using HeuristicLab.Data;
using HeuristicLab.MainForm;
using HeuristicLab.Optimization;

namespace HeuristicLab.Analysis.Statistics {
  [View("Correlations")]
  [Content(typeof(RunCollection), false)]
  public sealed partial class CorrelationView : ItemView {
    private enum ResultParameterType {
      Result,
      Parameter
    }

    public new RunCollection Content {
      get { return (RunCollection)base.Content; }
      set { base.Content = value; }
    }

    public override bool ReadOnly {
      get { return true; }
      set { /*not needed because results are always readonly */}
    }

    private Dictionary<string, ResultParameterType> resultsParameters;
    private bool suppressUpdates = false;

    public CorrelationView() {
      InitializeComponent();
      stringConvertibleMatrixView.Minimum = -1.0;
      stringConvertibleMatrixView.Maximum = 1.0;
    }

    protected override void OnContentChanged() {
      base.OnContentChanged();
      resultComboBox.Items.Clear();

      if (Content != null) {
        UpdateResultComboBox();
      }
      UpdateCaption();
    }

    private void UpdateCaption() {
      Caption = Content != null ? Content.OptimizerName + " Correlations" : ViewAttribute.GetViewName(GetType());
    }

    #region events
    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      Content.ItemsAdded += new HeuristicLab.Collections.CollectionItemsChangedEventHandler<IRun>(Content_ItemsAdded);
      Content.ItemsRemoved += new HeuristicLab.Collections.CollectionItemsChangedEventHandler<IRun>(Content_ItemsRemoved);
      Content.CollectionReset += new HeuristicLab.Collections.CollectionItemsChangedEventHandler<IRun>(Content_CollectionReset);
      Content.UpdateOfRunsInProgressChanged += Content_UpdateOfRunsInProgressChanged;
      RegisterRunEvents(Content);
    }

    protected override void DeregisterContentEvents() {
      base.DeregisterContentEvents();
      Content.ItemsAdded -= new HeuristicLab.Collections.CollectionItemsChangedEventHandler<IRun>(Content_ItemsAdded);
      Content.ItemsRemoved -= new HeuristicLab.Collections.CollectionItemsChangedEventHandler<IRun>(Content_ItemsRemoved);
      Content.CollectionReset -= new HeuristicLab.Collections.CollectionItemsChangedEventHandler<IRun>(Content_CollectionReset);
      Content.UpdateOfRunsInProgressChanged -= Content_UpdateOfRunsInProgressChanged;
      DeregisterRunEvents(Content);
    }

    private void RegisterRunEvents(IEnumerable<IRun> runs) {
      foreach (IRun run in runs) {
        RegisterRunResultsEvents(run);
        RegisterRunParametersEvents(run);
        run.PropertyChanged += run_PropertyChanged;
      }
    }

    private void DeregisterRunEvents(IEnumerable<IRun> runs) {
      foreach (IRun run in runs) {
        DeregisterRunResultsEvents(run);
        DeregisterRunParametersEvents(run);
        run.PropertyChanged -= run_PropertyChanged;
      }
    }

    private void RegisterRunResultsEvents(IRun run) {
      IObservableDictionary<string, IItem> dict = run.Results;
      dict.ItemsAdded += run_Changed;
      dict.ItemsRemoved += run_Changed;
      dict.ItemsReplaced += run_Changed;
      dict.CollectionReset += run_Changed;
    }

    private void DeregisterRunResultsEvents(IRun run) {
      IObservableDictionary<string, IItem> dict = run.Results;
      dict.ItemsAdded -= run_Changed;
      dict.ItemsRemoved -= run_Changed;
      dict.ItemsReplaced -= run_Changed;
      dict.CollectionReset -= run_Changed;
    }

    private void RegisterRunParametersEvents(IRun run) {
      IObservableDictionary<string, IItem> dict = run.Parameters;
      dict.ItemsAdded += run_Changed;
      dict.ItemsRemoved += run_Changed;
      dict.ItemsReplaced += run_Changed;
      dict.CollectionReset += run_Changed;
    }

    private void DeregisterRunParametersEvents(IRun run) {
      IObservableDictionary<string, IItem> dict = run.Parameters;
      dict.ItemsAdded -= run_Changed;
      dict.ItemsRemoved -= run_Changed;
      dict.ItemsReplaced -= run_Changed;
      dict.CollectionReset -= run_Changed;
    }

    private void Content_CollectionReset(object sender, HeuristicLab.Collections.CollectionItemsChangedEventArgs<IRun> e) {
      DeregisterRunEvents(e.OldItems);
      RegisterRunEvents(e.Items);
      UpdateUI();
    }

    private void Content_ItemsRemoved(object sender, HeuristicLab.Collections.CollectionItemsChangedEventArgs<IRun> e) {
      DeregisterRunEvents(e.Items);
      UpdateUI();
    }

    private void Content_ItemsAdded(object sender, HeuristicLab.Collections.CollectionItemsChangedEventArgs<IRun> e) {
      RegisterRunEvents(e.Items);
      UpdateUI();
    }

    private void Content_UpdateOfRunsInProgressChanged(object sender, EventArgs e) {
      suppressUpdates = Content.UpdateOfRunsInProgress;
      UpdateUI();
    }

    private void run_Changed(object sender, EventArgs e) {
      if (InvokeRequired) {
        Invoke(new EventHandler(run_Changed), sender, e);
      }
      UpdateUI();
    }

    void run_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e) {
      UpdateUI();
    }
    #endregion

    private void UpdateUI() {
      if (!suppressUpdates) {
        UpdateResultComboBox();
        RebuildCorrelationTable();
      }
    }

    private void UpdateResultComboBox() {
      resultsParameters = GetRowNames();
      string selectedResult = (string)this.resultComboBox.SelectedItem;
      resultComboBox.Items.Clear();
      resultComboBox.Items.AddRange(resultsParameters.Keys.ToArray());

      if (selectedResult != null && resultComboBox.Items.Contains(selectedResult)) {
        resultComboBox.SelectedItem = selectedResult;
      } else if (resultComboBox.Items.Count > 0) {
        resultComboBox.SelectedItem = resultComboBox.Items[0];
      }
    }

    private List<string> GetResultRowNames() {
      var results = (from run in Content
                     where run.Visible
                     from result in run.Results
                     where result.Value is DoubleValue || result.Value is IntValue
                     select result.Key).Distinct().ToList();

      return results;
    }

    private List<string> GetParameterRowNames() {
      var parameters = (from run in Content
                        where run.Visible
                        from parameter in run.Parameters
                        where parameter.Value is DoubleValue || parameter.Value is IntValue
                        select parameter.Key).Distinct().ToList();

      return parameters;
    }

    private Dictionary<string, ResultParameterType> GetRowNames() {
      Dictionary<string, ResultParameterType> ret = new Dictionary<string, ResultParameterType>();

      var results = GetResultRowNames();
      var parameters = GetParameterRowNames();

      foreach (var r in results) {
        ret.Add(r, ResultParameterType.Result);
      }
      foreach (var p in parameters) {
        if (!ret.ContainsKey(p)) {
          ret.Add(p, ResultParameterType.Parameter);
        }
      }

      return ret;
    }

    private List<double> GetDoublesFromResults(List<IRun> runs, string key) {
      List<double> res = new List<double>();

      foreach (var r in runs) {
        if (r.Results[key] is DoubleValue) {
          res.Add(((DoubleValue)r.Results[key]).Value);
        } else {
          res.Add(((IntValue)r.Results[key]).Value);
        }
      }
      return res;
    }

    private List<double> GetDoublesFromParameters(List<IRun> runs, string key) {
      List<double> res = new List<double>();

      foreach (var r in runs) {
        if (r.Parameters[key] is DoubleValue) {
          res.Add(((DoubleValue)r.Parameters[key]).Value);
        } else {
          res.Add(((IntValue)r.Parameters[key]).Value);
        }
      }
      return res;
    }

    private void RebuildCorrelationTable() {
      string resultName = (string)resultComboBox.SelectedItem;

      var columnNames = new string[2];
      columnNames[0] = "Pearson product-moment correlation coefficient";
      columnNames[1] = "Spearman's rank correlation coefficient";

      var runs = Content.Where(x => x.Results.ContainsKey(resultName) && x.Visible);

      DoubleMatrix dt = new DoubleMatrix(resultsParameters.Count(), columnNames.Count());
      dt.RowNames = resultsParameters.Keys.ToArray();
      dt.ColumnNames = columnNames;

      int j = 0;
      foreach (var rowName in resultsParameters) {
        var resultVals = GetDoublesFromResults(runs.Where(x => x.Results.ContainsKey(resultName)).Where(x => x.Results[resultName] is DoubleValue || x.Results[resultName] is IntValue).ToList(), resultName);

        List<double> resultRowVals;
        if (rowName.Value == ResultParameterType.Result) {
          resultRowVals = GetDoublesFromResults(runs.Where(x => x.Results.ContainsKey(rowName.Key)).Where(x => x.Results[rowName.Key] is DoubleValue || x.Results[rowName.Key] is IntValue).ToList(), rowName.Key);
        } else {
          resultRowVals = GetDoublesFromParameters(runs.Where(x => x.Parameters.ContainsKey(rowName.Key)).Where(x => x.Parameters[rowName.Key] is DoubleValue || x.Parameters[rowName.Key] is IntValue).ToList(), rowName.Key);
        }

        resultVals = resultVals.Where(x => !double.IsNaN(x) && !double.IsNegativeInfinity(x) && !double.IsPositiveInfinity(x)).ToList();
        resultRowVals = resultRowVals.Where(x => !double.IsNaN(x) && !double.IsNegativeInfinity(x) && !double.IsPositiveInfinity(x)).ToList();

        if (resultRowVals.Count == 0 || resultVals.Count == 0 || resultRowVals.Count != resultVals.Count) {
          dt[j, 0] = double.NaN;
          dt[j++, 1] = double.NaN;
        } else {
          dt[j, 0] = alglib.pearsoncorr2(resultVals.ToArray(), resultRowVals.ToArray());
          dt[j++, 1] = alglib.spearmancorr2(resultVals.ToArray(), resultRowVals.ToArray());
        }
      }

      dt.SortableView = true;
      stringConvertibleMatrixView.Content = dt;
    }

    private void resultComboBox_SelectedIndexChanged(object sender, EventArgs e) {
      RebuildCorrelationTable();
    }
  }
}
